%{

#include"parse_config.h"

extern FILE *config_yyin;
extern int config_yylineno;
extern int config_yylex(void);

extern int config_count_column, config_count_line;
extern char config_count_buffer[MAXSTRINGSIZE];

int errorcode;

#ifdef CONF_YYCLEAN
void config_yyclean(void);
#endif

void config_yyerror(const char *s);

%}

%union {
  int number;
  char *string;
}

%token <string> TOKEN_STRING TOKEN_VARIABLE TOKEN_ID
%token <number> TOKEN_INT
%token <number> '=' '(' ')' ','
%token TOKEN_EOL TOKEN_NULL
%token TOKEN_HOME TOKEN_SUBST TOKEN_SUBSTANDSTAT TOKEN_CONCAT TOKEN_CONCATNOTNULL

%type <string> lines line rvalue concat_rvalue_list concatnotnull_rvalue_list

%start lines
%%

lines
  : line
  | line TOKEN_EOL lines
  ;

line
  : /* NULL */ { }
  | TOKEN_VARIABLE '=' rvalue
    {
      setvar($1, $3);

      if(!strcmp($1, "%WRAPPER_USER"))	/* cache user info */
      {
      	struct passwd* user;
	
	user=getpwnam($3);
	if(user != NULL)
	{
	  char tmp[MAXSTRINGSIZE];

	  snprintf(tmp, MAXSTRINGSIZE, "%d", user->pw_uid);
    #ifdef HAVE_SNPRINTF_NOTTERMINATED
	  tmp[MAXSTRINGSIZE - 1] = '\0';
    #endif
	  setvar("%WRAPPER_UID", tmp);

	  snprintf(tmp, MAXSTRINGSIZE, "%d", user->pw_gid);
    #ifdef HAVE_SNPRINTF_NOTTERMINATED
	  tmp[MAXSTRINGSIZE - 1] = '\0';
    #endif
	  setvar("%WRAPPER_GID", tmp);

	  setvar("%WRAPPER_HOME", user->pw_dir);

	  setvar("%WRAPPER_SHELL", user->pw_shell);
	}
	else
	{
	  setvar("%WRAPPER_UID", NULL);
	  setvar("%WRAPPER_GID", NULL);
	  setvar("%WRAPPER_HOME", NULL);
	  setvar("%WRAPPER_SHELL", NULL);
	}
      }

      if($1 != NULL)
      {
	free($1);
      }
      if($3 != NULL)
      {
	free($3);
      }
    }
  ;

rvalue
  : TOKEN_STRING
    {
      $$ = $1;
    }
  | TOKEN_INT
    {
      char *tmp;

      tmp = (char *)malloc(MAXSTRINGSIZE * sizeof(char));

      snprintf(tmp, MAXSTRINGSIZE, "%d", $1);
    #ifdef HAVE_SNPRINTF_NOTTERMINATED
      tmp[MAXSTRINGSIZE - 1] = '\0';
    #endif

      $$ = tmp;
    }
  | TOKEN_VARIABLE
    {
      char *tmp;

      if(getvar($1) != NULL)
      {
	tmp = (char *)malloc((strlen(getvar($1)) + 1) + sizeof(char));

	strcpy(tmp, getvar($1));

	$$ = tmp;
      }
      else
      {
	$$ = NULL;
      }

      if($1 != NULL)
      {
	free($1);
      }
    }
  | TOKEN_NULL
    {
      $$ = NULL;
    }
  | TOKEN_HOME '(' rvalue ')'
    {
      char *ptr;

      ptr = (char *)malloc(MAXSTRINGSIZE * sizeof(char));

      if(home($3, getvar("%WRAPPER_USER"), getvar("%WRAPPER_HOME"), ptr, MAXSTRINGSIZE))
      {
	free(ptr);
	$$ = NULL;
      }
      else
      {
	$$ = ptr;
      }

      if($3 != NULL)
      {
	free($3);
      }
    }
  | TOKEN_SUBST '(' rvalue ',' rvalue ',' rvalue ')'
    {
      char *ptr;

      ptr = (char *)malloc(MAXSTRINGSIZE * sizeof(char));

      if(subst($3, $5, $7, ptr, MAXSTRINGSIZE))
      {
	free(ptr);
	$$ = NULL;
      }
      else
      {
	$$ = ptr;
      }

      if($3 != NULL)
      {
	free($3);
      }
      if($5 != NULL)
      {
	free($5);
      }
      if($7 != NULL)
      {
	free($7);
      }
    }
  | TOKEN_SUBSTANDSTAT '(' rvalue ',' rvalue ',' rvalue ')'
    {
      char *ptr;

      ptr = (char *)malloc(MAXSTRINGSIZE * sizeof(char));

      if(substandstat($3, $5, $7, ptr, MAXSTRINGSIZE))
      {
	free(ptr);
	$$ = NULL;
      }
      else
      {
	$$ = ptr;
      }

      if($3 != NULL)
      {
	free($3);
      }
      if($5 != NULL)
      {
	free($5);
      }
      if($7 != NULL)
      {
	free($7);
      }
    }
  | TOKEN_CONCAT '(' concat_rvalue_list ')'
    {
      $$ = $3;
    }
  | TOKEN_CONCATNOTNULL '(' concatnotnull_rvalue_list ')'
    {
      $$ = $3;
    }
  ;

concat_rvalue_list
  : rvalue
    {
      $$ = $1;
    }
  | rvalue ',' concat_rvalue_list
    {
      char *ptr;

      ptr = (char *)malloc(MAXSTRINGSIZE * sizeof(char));

      if(concat($1, $3, ptr, MAXSTRINGSIZE))
      {
	free(ptr);
	$$ = NULL;
      }
      else
      {
	$$ = ptr;
      }

      if($1 != NULL)
      {
	free($1);
      }
      if($3 != NULL)
      {
	free($3);
      }
    }
  ;

concatnotnull_rvalue_list
  : rvalue
    {
      $$ = $1;
    }
  | rvalue ',' concatnotnull_rvalue_list
    {
      char *ptr;

      ptr = (char *)malloc(MAXSTRINGSIZE * sizeof(char));

      if(concatnotnull($1, $3, ptr, MAXSTRINGSIZE))
      {
	free(ptr);
	$$ = NULL;
      }
      else
      {
	$$ = ptr;
      }

      if($1 != NULL)
      {
	free($1);
      }
      if($3 != NULL)
      {
	free($3);
      }
    }
  ;

%%

void config_yyerror(const char *s)
{
  char tmp1[MAXSTRINGSIZE];
  char tmp2[MAXSTRINGSIZE];

  snprintf(tmp1, MAXSTRINGSIZE, "modiwrap.conf:%d: error: %s",
	   config_count_line, s);
#ifdef HAVE_SNPRINTF_NOTTERMINATED
  tmp1[MAXSTRINGSIZE - 1] = '\0';
#endif

  snprintf(tmp2, MAXSTRINGSIZE, "%s", config_count_buffer);
#ifdef HAVE_SNPRINTF_NOTTERMINATED
  tmp2[MAXSTRINGSIZE - 1] = '\0';
#endif

  wraperrorlog(tmp1, tmp2);

  errorcode = -1;
}

int parse_config(const char *filename)
{
  FILE *conf;

  if(filename == NULL)
  {
    return (-1);
  }

  conf = fopen(filename, "r");
  if(conf == NULL)
  {
    wraperrorlog(filename, strerror(errno));
    return (-1);
  }

  config_yyin = conf;

  errorcode = 0;
  config_yyparse();

  fclose(conf);

#ifdef CONF_YYCLEAN
  config_yyclean();
#endif

  return (errorcode);
}
