%{

#include"htaccess.h"
#include"parse_htaccess.h"

extern int nphmodiwrap;

extern FILE *htaccess_yyin;
extern int htaccess_yylineno;
extern int htaccess_yylex(void);

extern int htaccess_count_column, htaccess_count_line;
extern char htaccess_count_buffer[MAXSTRINGSIZE];

const char *script_filename;
const char *document_root;
const char *request_method;
const char *http_authorization_type;
const char *http_user;
const char *http_password;

int errorcode;
int authorized;

char *authname = NULL;
char *authtype = NULL;
char *authgroupfile = NULL;
char *authuserfile = NULL;

void htaccess_yyerror(const char *s);

%}

%union {
  int number;
  char *string;
}

%token <string> TOKEN_STRING
%token TOKEN_EOL
%token TOKEN_AUTHORITATIVE
%token TOKEN_AUTHNAME
%token TOKEN_AUTHTYPE
%token TOKEN_REQUIRE
%token TOKEN_REQUIREUSER TOKEN_REQUIREGROUP TOKEN_REQUIREVALIDUSER
%token TOKEN_AUTHGROUPFILE
%token TOKEN_AUTHUSERFILE
%token TOKEN_ERRORDOCUMENT

%type <string> lines line
%type <number> user_list group_list

%start lines
%%

lines
  : line
  | line TOKEN_EOL lines
  ;

line
  : /* NULL */ { }
  | TOKEN_AUTHORITATIVE TOKEN_STRING
    {
      if($2 != NULL)
      {
	free($2);
      }
    }
  | TOKEN_AUTHNAME TOKEN_STRING
    {
      if(authname != NULL)
      {
	free(authname);
      }

      if($2 != NULL)
      {
	authname = (char *)malloc((strlen($2) + 1) * sizeof(char));
	strcpy(authname, $2);
	free($2);
      }
      else
      {
	authname = NULL;
      }
    }
  | TOKEN_AUTHTYPE TOKEN_STRING
    {
      if(authtype != NULL)
      {
	free(authtype);
      }

      if($2 != NULL)
      {
	authtype = (char *)malloc((strlen($2) + 1) * sizeof(char));

	strcpy(authtype, $2);

	if(http_authorization_type != NULL && strcmp(authtype, http_authorization_type))	/* different authtypes */
	{
	  return (-1);		/* quits parsing */
	}

	free($2);
      }
      else
      {
	authtype = NULL;
      }
    }
  | TOKEN_REQUIRE TOKEN_REQUIREUSER user_list
    {
      if($3)			/* failed */
      {
	authorized = 0;
	return (-1);
      }
    }
  | TOKEN_REQUIRE TOKEN_REQUIREGROUP group_list
    {
      if($3)			/* failed */
      {
	authorized = 0;
	return (-1);
      }
    }
  | TOKEN_REQUIRE TOKEN_REQUIREVALIDUSER
    {
      if(htaccesscheckuser(http_user, http_password, authuserfile, authtype))	/* failed */
      {
	authorized = 0;
	return (-1);
      }
    }
  | TOKEN_AUTHGROUPFILE TOKEN_STRING
    {
      if(authgroupfile != NULL)
      {
	free(authgroupfile);
	authgroupfile = NULL;
      }

      if($2 != NULL)
      {
	if(strstr($2, "../") == NULL)
	{
	  if($2[0] == '/')		/* absolute path */
	  {
    #ifdef CONF_ENFORCE_DOCUMENT_ROOT
	    if(!strncmp($2, document_root, strlen(document_root)))
	    {
    #endif
    #ifdef CONF_ENFORCE_NOT_ETC
	      if(strncmp($2, "/etc/", 5))
	      {
    #endif
		authgroupfile = (char *)malloc((strlen($2) + 1) * sizeof(char));

		strcpy(authgroupfile, $2);
    #ifdef CONF_ENFORCE_NOT_ETC
	      }
    #endif
    #ifdef CONF_ENFORCE_DOCUMENT_ROOT
	    }
    #endif
	  }
	  else
	  {
	    authgroupfile =
	      (char *)malloc((strlen(document_root) + 1 + strlen($2) + 1) *
			     sizeof(char));
	    strcpy(authgroupfile, document_root);
	    if(strlen(authgroupfile) < 1
	       || authgroupfile[strlen(authgroupfile) - 1] != '/')
	    {
	      strcat(authgroupfile, "/");
	    }
	    strcat(authgroupfile, $2);
	  }
	}

	free($2);
      }
      else
      {
	authgroupfile = NULL;
      }
    }
  | TOKEN_AUTHUSERFILE TOKEN_STRING
    {
      if(authuserfile != NULL)
      {
	free(authuserfile);
	authuserfile = NULL;
      }

      if($2 != NULL)
      {
	if(strstr($2, "../") == NULL)
	{
	  if($2[0] == '/')		/* absolute path */
	  {
    #ifdef CONF_ENFORCE_DOCUMENT_ROOT
	    if(!strncmp($2, document_root, strlen(document_root)))
	    {
    #endif
    #ifdef CONF_ENFORCE_NOT_ETC
	      if(strncmp($2, "/etc/", 5))
	      {
    #endif
		authuserfile = (char *)malloc((strlen($2) + 1) * sizeof(char));

		strcpy(authuserfile, $2);
    #ifdef CONF_ENFORCE_NOT_ETC
	      }
    #endif
    #ifdef CONF_ENFORCE_DOCUMENT_ROOT
	    }
    #endif
	  }
	  else
	  {
	    authuserfile =
	      (char *)malloc((strlen(document_root) + 1 + strlen($2) + 1) *
			     sizeof(char));

	    strcpy(authuserfile, document_root);
	    if(strlen(authuserfile) < 1
	       || authuserfile[strlen(authuserfile) - 1] != '/')
	    {
	      strcat(authuserfile, "/");
	    }
	    strcat(authuserfile, $2);
	  }
	}

	free($2);
      }
    }
  | TOKEN_ERRORDOCUMENT TOKEN_STRING TOKEN_STRING
    {
    #ifdef CONF_USE_ERRORDOCUMENT
      if($2 != NULL)
      {
	if(atoi($2) != 0)
	{
	  registererrorpage(atoi($2), $3);
	}
	free($2);

	if($3 != NULL)
	{
	  free($3);
	}
      }
    #endif
    }
  ;

user_list
  : TOKEN_STRING
    {
      if($1 != NULL)
      {
	if(!strcmp($1, http_user))
	{
	  $$ =
	    htaccesscheckuser(http_user, http_password, authuserfile, authtype);
	}
	else
	{
	  $$ = 1;
	}

	free($1);
      }
    }
  | TOKEN_STRING user_list
    {
      if($1 != NULL)
      {
	if(!strcmp($1, http_user))
	{
	  $$ =
	    htaccesscheckuser(http_user, http_password, authuserfile, authtype);
	}
	else
	{
	  $$ = $2;
	}
      }

      free($1);
    }
  ;

group_list
  : TOKEN_STRING
    {
      if($1 != NULL)
      {
	if(!htaccessuseringroup(http_user, $1, authgroupfile))	/* yes */
	{
	  $$ =
	    htaccesscheckuser(http_user, http_password, authuserfile, authtype);
	}
	else
	{
	  $$ = 1;
	}

	free($1);
      }
    }
  | TOKEN_STRING group_list
    {
      if($1 != NULL)
      {
	if(!htaccessuseringroup(http_user, $1, authgroupfile))	/* yes */
	{
	  $$ =
	    htaccesscheckuser(http_user, http_password, authuserfile, authtype);
	}
	else
	{
	  $$ = $2;
	}
      }

      free($1);
    }
  ;

%%

void htaccess_yyerror(const char *s)
{
  char tmp1[MAXSTRINGSIZE];
  char tmp2[MAXSTRINGSIZE];

  snprintf(tmp1, MAXSTRINGSIZE, ".htaccess:%d: error: %s",
	   htaccess_count_line, s);
#ifdef HAVE_SNPRINTF_NOTTERMINATED
  tmp1[MAXSTRINGSIZE - 1] = '\0';
#endif

  snprintf(tmp2, MAXSTRINGSIZE, "%s", htaccess_count_buffer);
#ifdef HAVE_SNPRINTF_NOTTERMINATED
  tmp2[MAXSTRINGSIZE - 1] = '\0';
#endif

  wraperrorlog(tmp1, tmp2);

  errorcode = -1;
}

int parse_htaccess(const char *htaccess, const char *local_script_filename,
		   const char *local_document_root,
		   const char *local_request_method,
		   const char *local_http_authorization_type,
		   const char *local_http_user,
		   const char *local_http_password)
{
  FILE *ht;

  if(htaccess == NULL || local_request_method == NULL)
  {
    return (-1);
  }

  ht = fopen(htaccess, "r");
  if(ht == NULL)
  {
    wraperrorlog(htaccess, strerror(errno));
    return (-1);
  }

  htaccess_yyin = ht;

  script_filename = local_script_filename;
  document_root = local_document_root;
  request_method = local_request_method;
  http_authorization_type = local_http_authorization_type;
  http_user = local_http_user;
  http_password = local_http_password;

  authorized = 1;
  errorcode = 0;
  htaccess_yyparse();

  fclose(ht);

  if(errorcode == 0 && authorized == 0)
  {
    if(authname == NULL)
    {
      authname = "";
    }

    if(authtype == NULL)
    {
      authtype = "Basic";
    }

    wrapaskforauth(authtype, authname);
    errorcode = 1;
  }

  if(errorcode == 0 && authorized)
  {
    setvar("$AUTH_TYPE", local_http_authorization_type);
    setvar("$REMOTE_USER", local_http_user);
  }

  if(authname != NULL)
  {
    free(authname);
    authname = NULL;
  }
  if(authtype != NULL)
  {
    free(authtype);
    authtype = NULL;
  }
  if(authgroupfile != NULL)
  {
    free(authgroupfile);
    authgroupfile = NULL;
  }
  if(authuserfile != NULL)
  {
    free(authuserfile);
    authuserfile = NULL;
  }

  return (errorcode);
}
